#!/usr/bin/env python3
# automome.py
# automome - the Ottish AUTOMEME clone
# Copyright (C) 2014  Penguin Development
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import getopt
import itertools
import os
import os.path
import random
import re
import sys

class Replaceable(object):
    """Class for which a randomly picked word is substituted."""
    scanned = False
    scandir = '.'
    classes = {}

    def __init__(self, cls, default=None, prob=1.0, **kwargs):
        """Constructor. `cls' is the word class that this Replaceable should
        belong to; it should match the file name of a .dict.txt file in
        Replaceable.scandir. `cls' may also be a tuple or list of multiple
        word classes. `prob' is the probability of a replacement being chosen;
        (1-prob) is the probaility of `default' being used instead. If
        `default' evaluates to False, `prob' has no effect."""
        self.cls = cls
        self.default = default
        self.prob = prob
        self.kwargs = kwargs
        self.scanclasses()
        if type(self.cls) == str:
            self.cls = (self.cls,)
        for c in self.cls:
            if c not in self.classes:
                raise KeyError(
                        'Mustard: word class `{}\' not found!'.format(c))

    @classmethod
    def scanclasses(cls, force = False):
        """Scans the directory Replaceable.scandir for files named *.dict.txt.
        The files' base names (i.e. the part before .dict.txt) are used as
        names for word classes; the files contain one word per line,
        whitespace is stripped off the start and end of each line, and empty
        lines (after stripping) are ignored.

        This method aborts immediately if it has been called previously,
        unless `force' evaluates to True."""
        if cls.scanned and not force:
            return

        for fn in os.listdir(cls.scandir):
            ffn = os.path.join(cls.scandir, fn)
            if fn.endswith('.dict.txt') and os.path.isfile(ffn):
                f = open(ffn, 'r', encoding='UTF-8')
                # the encoding in the line above is to prevent CGI mustard
                data = [x.strip() for x in f.read().splitlines()]
                f.close()
                data = [d for d in data if d]
                if len(data):
                    cls.classes[fn[:-9]] = data

        cls.scanned = True

    def replace(self):
        """Picks a random word from the word class(es) self.cls that will be
        used in lieu of this Replaceable."""
        repl = ''
        if self.default and random.random() > self.prob:
            repl = self.default
        else:
            repl = random.choice(tuple(itertools.chain.from_iterable(
                [self.classes[c] for c in self.cls])))

        return repl

class IndexMap(object):
    """Class that applies a function to the replacement, given the argument
    index of that replacement. E.g. let R = ('foo', IndexMap(0, f)) be the
    replacement tuple, then this will get expanded to ('foo', f('foo'))."""
    def __init__(self, idx, f = lambda x: x):
        self.index = idx
        self.func = f

    def apply(self, data):
        return self.func(data)

class Meme(object):
    """Meme class."""
    def __init__(self, template, *args):
        self.template = template
        self.args = args

    def __str__(self):
        """Returns a fully-filled version of the template."""
        args = [x.replace() if isinstance(x, Replaceable) else x for x in
                self.args]
        for i in range(len(args)):
            if isinstance(args[i], IndexMap):
                args[i] = args[i].apply(args[args[i].index]) # ARGCEPTION

        try:
            return self.template.format(*args)
        except:
            print(">>> In template " + self.template, file=sys.stderr)
            raise


# Praparatory block to set the scan dir and seed the PRNG if automome.py is
# called directly.
if __name__ == '__main__':
    p = os.getenv('AUTOMOME_PATH')
    if (p):
        Replaceable.scandir = p
    elif not Replaceable.scandir:
        Replaceable.scandir = os.path.abspath(os.path.dirname(sys.argv[0]))

    random.seed()

R = lambda *x: Replaceable(x)
Rd = lambda d, p, *x: Replaceable(x, d, p)
skipword = lambda x: ' '.join(x.split(' ')[1:])

def skipuntilvowel(s):
    mat = re.compile(r'[AEIOU]').search(s)
    if mat:
        return s[mat.start():]
    else:
        return s

memes = [
    Meme('ALL YOUR {} ARE BELONG TO {}',
        R('noun-singular'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('ATTACK ITS WEAK POINT FOR {} {}',
        R('adjective'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('BECAUSE YOU ARE _{} {}_',
        Rd('HEADCRAB', 0.33, 'adjective'),
        R('noun-singular', 'noun-mass')),
    Meme('BETTER {} MY OWN {}',
        R('verb-tr-inf'),
        R('noun-singular', 'noun-mass')),
    Meme('BLESS YOU, {}. BL{}.',
        R('noun-plural', 'noun-mass', 'character'),
        IndexMap(0, skipuntilvowel)),
    Meme('CH*RPING {}, HOW DO THEY WORK?',
        R('noun-plural')),
    Meme('DAMNIT JIM, I\'M {}, NOT {}',
        R('noun-singular-article'),
        R('noun-singular-article')),
    Meme('DISREGARD THAT, I {} {}',
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('DR. {}, OR HOW I LEARNED TO STOP {} AND LOVE THE {}',
        R('noun-singular', 'noun-mass', 'character'),
        Rd('WORRYING', 0.1, 'verb-intr-cont'),
        R('noun-singular')),
    Meme('GREAT LORD RANDALL IT\'S {} GET IN THE {}',
        R('noun-singular-article'),
        Rd('CAR', 0.33, 'noun-singular', 'noun-mass')),
    Meme('HELLO. MY NAME IS {}. YOU {} MY {}. PREPARE TO {}.',
        R('character'),
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass'),
        R('verb-intr-inf')),
    Meme('HELLO? YES, THIS IS {}',
        R('noun-singular', 'noun-mass', 'character')),
    Meme('HOW DO I {} {}?',
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('HOW IS {} FORMED? HOW {} GET {}?',
        R('noun-singular', 'noun-mass'),
        R('noun-singular', 'noun-mass', 'character'),
        Rd('PRAGNENT', 0.33, 'adjective')),
    Meme('I ACCIDENTALLY {}, IS THIS BAD?',
        R('noun-singular-article')),
    Meme('I AM {}, HEAR ME {}',
        R('character'),
        Rd('ROAR', 0.5, 'verb-intr-inf')),
    Meme('I DON\'T ALWAYS {}, BUT WHEN I DO, I {} {}',
        R('verb-intr-inf'),
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('I, FOR ONE, WELCOME OUR NEW {} OVERLORDS',
        R('noun-singular')),
    Meme('I HAS {}. YOU CAN HAS MY {}.',
        R('noun-singular-article'),
        IndexMap(0, skipword)),
    Meme('I HAVE REPORTED YOU TO {} FOR {} {} AS IT IS A CRIME',
        R('character'),
        Rd('STEALING', 0.25, 'verb-tr-cont'),
        R('noun-plural', 'noun-mass')),
    Meme('I LIKE MY {} LIKE I LIKE MY {}: COVERED IN {}',
        Rd('WOMEN', 0.5, 'noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('I MADE YOU {} BUT I {} IT',
        R('noun-singular-article'),
        R('verb-tr-past')),
    Meme('IMMA CHARGIN MAH {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('I\'M IN UR {}, {} UR {}',
        R('noun-singular', 'noun-mass'),
        R('verb-tr-cont'),
        R('noun-plural', 'noun-mass')),
    Meme('IN SOVIET RUSSIA, {} {} _YOU_',
        R('noun-singular', 'noun-mass', 'character'),
        R('verb-tr-present')),
    Meme('I SEE WHAT YOU {} THERE',
        R('verb-tr-past')),
    Meme('IS IT CAN BE {} TIEM NAO PLEES?',
        R('noun-singular')),
    Meme('I THINK {} IS A PRETTY {} {}. '
            'EH {} {} AND DOESN\'T AFRAID OF ANYTHING',
        R('character'),
        Rd('TREEISH', 0.2, 'adjective'),
        Rd('GUY', 0.1, 'noun-singular'),
        R('verb-tr-present'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('I {} _ALL_ THE {}',
        R('verb-tr-past'),
        R('noun-plural', 'noun-mass')),
    Meme('I {}, THEREFORE I {}',
        Rd('THINK', 0.5, 'verb-intr-inf'),
        R('verb-intr-inf')),
    Meme('I WILL NOT {} THIS {} -- IT IS {}',
        R('verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular-article', 'noun-mass', 'adjective', 'character')),
    Meme('MAXIMUM {} YIELDS MAXIMUM {}',
        R('noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('MY {} IS FULL OF {}',
        R('noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('MY {} JUST {} ME; SHIT WAS SO {}',
        R('noun-singular', 'noun-mass'),
        R('verb-tr-past'),
        Rd('CASH', 0.5, 'adjective')),
    Meme('MY {}, LET ME SHOW YOU IT',
        R('noun-singular', 'noun-mass')),
    Meme('NOT SURE IF {} OR JUST {}',
        R('verb-intr-cont'),
        R('verb-intr-cont', 'adjective')),
    Meme('NOOOO THEY BE {} MY {}',
        R('verb-tr-cont'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('NO, {}, YOU ARE THE {}',
        R('noun-singular', 'noun-mass', 'character'),
        R('noun-singular', 'noun-mass')),
    Meme('OH GLR HOW DID {} GET HERE I AM NOT GOOD WITH {}',
        Rd('THIS', 0.5, 'noun-plural', 'noun-mass', 'character'),
        R('noun-plural', 'noun-mass')),
    Meme('OH HAI, I {} YOUR {}',
        R('verb-tr-past'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('PROTIP: TO {} THE {}, {} IT UNTIL IT {}',
        Rd('DEFEAT', 0.75, 'verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('verb-tr-inf'),
        R('verb-intr-present')),
    Meme('PUSH BUTAN, RECEIVE {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('PUSH BUTTON, RECEIVE {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('SCREAMING {}: {} DOESN\'T LIKE IT',
        R('noun-singular', 'character'),
        R('character')),
    Meme('SO I HERD U {} {}',
        Rd('LIEK', 0.25, 'verb-tr-inf'),
        R('noun-plural', 'noun-mass')),
    Meme('THANKS, {}. TH{}.',
        R('noun-plural', 'noun-mass', 'character'),
        IndexMap(0, skipuntilvowel)),
    Meme('THANK YOU {}, BUT OUR {} IS IN ANOTHER {}',
        R('character'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular')),
    Meme('THAT\'S NOT {0}; _THIS_ IS {0}',
        R('noun-singular-article')),
    Meme('THE {} IS A LIE',
        R('noun-singular')),
    Meme('THERE IS {} IN MY {}, YOUR ARGUMENT IS INVALID',
        R('noun-singular-article'),
        Rd('BEARD', 0.67, 'noun-singular', 'noun-mass')),
    Meme('THEY SEE ME {}, THEY {}',
        R('verb-intr-cont'),
        Rd('HATIN', 0.67, 'verb-intr-cont')),
    Meme('"THIS IS NOT ACCEPTABLE!" I SCREAMED AS {} {}',
        R('character'),
        R('verb-intr-past')),
    Meme('THIS IS YOUR {} ON {}',
        Rd('BRAIN', 0.25, 'noun-singular'),
        R('noun-plural', 'noun-mass')),
    Meme('THIS IS {}. I CAN TELL BY THE {}, '
            'AND FROM HAVING SEEN A LOT OF {} IN MY DAY',
        Rd('PHOTOSHOPPED', 0.67, 'adjective'),
        R('noun-singular', 'noun-plural', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('VERY IMPORTANT FOR {} TO {} HIM {}',
        R('character'),
        Rd('DO', 0.75, 'verb-tr-inf'),
        Rd('JOB', 0.5, 'noun-singular', 'noun-mass')),
    Meme('WE\'VE SECRETLY REPLACED {}\'S {} WITH {}. '
            'LET\'S SEE IF HE NOTICES...',
        R('character'),
        R('noun-singular', 'noun-mass'),
        R('noun-singular-article')),
    Meme('WHAT _ARE_ {}? WE JUST DON\'T KNOW.',
        R('noun-plural')),
    Meme('WHATEVER {} YOUR {}',
        R('verb-tr-present'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('WHAT THE CH*RP? YOU ARE {}. AND {}',
        R('noun-singular', 'noun-mass', 'character'),
        R('adjective')),
    Meme('YOU HAVE {} MOST OF A RIVERISH {}. '
            'PLEASE PICK YOUR {} WITH GREATER CARE.',
        R('verb-tr-past'),
        R('noun-singular'),
        R('noun-plural', 'noun-mass')),
    Meme('YOU WOULDN\'T {} {}',
        Rd('DOWNLOAD', 0.5, 'verb-tr-inf'),
        R('noun-singular-article')),
    Meme('YO WAWLPY, I HEARD YOU LIKE {0} SO WE PUT SOME {0} IN YOUR {0} '
            'SO YOU CAN {1} WHILE YOU {1}',
        R('noun-plural', 'noun-mass'),
        R('verb-intr-inf')),
    Meme('YO {}, I\'M REALLY HAPPY FOR YOU, IMMA LET YOU FINISH, '
            'BUT {} HAD ONE OF THE BEST {} OF ALL TIME',
        R('character'),
        R('character'),
        R('noun-plural')),
    Meme('{} DON\'T KNOW ABOUT MY {}',
        R('noun-plural'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('{}? IN _MY_ {}?',
        R('noun-plural', 'noun-mass'),
        R('noun-singular', 'noun-mass')),
    Meme('{} IS {} TOO',
        R('noun-singular-article'),
        R('adjective')),
    Meme('{} IS SERIOUS {}',
        R('noun-singular', 'character'),
        R('noun-singular')),
    Meme('{} IS THE {} OF THE {}',
        R('noun-singular', 'character'),
        R('noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass')),
    Meme('{}\'S CLOSED DUE TO {}',
        Rd('POOL', 0.5, 'noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('{} OR GTFO',
        R('noun-plural', 'noun-mass')),
    Meme('{} THE {}... WITH YOUR {}',
        R('verb-tr-inf'),
        R('noun-plural', 'noun-mass'),
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('{} {} ARE _SUPERIOR_',
        R('adjective'),
        R('noun-plural'))
]

memes_unmolpish = [
    Meme('BITCHES DON\'T KNOW ABOUT MY {}',
        R('noun-singular', 'noun-plural', 'noun-mass')),
    Meme('HELLO. MY NAME IS {}. YOU KILLED MY {}. PREPARE TO {}.',
        R('character'),
        R('noun-singular', 'noun-plural', 'noun-mass'),
        Rd('DIE', 0.1, 'verb-intr-inf')),
    Meme('I\'M {0}! I\'M {0}! SUCK MY DIIIIICK! I\'M {0}!',
        R('noun-singular-article', 'character')),
    Meme('PROTIP: TO {} THE {}, {} IT UNTIL IT DIES',
        Rd('DEFEAT', 0.75, 'verb-tr-inf'),
        R('noun-singular', 'noun-mass'),
        R('verb-tr-inf')),
    Meme('{} HAET {}',
        Rd('LOLI', 0.5, 'character'),
        R('noun-singular', 'noun-plural', 'noun-mass', 'character')),
    Meme('{} IS PIG DISGUSTING',
        R('noun-singular', 'character')),
    Meme('{} IS THE {} THAT IS KILLING {}',
        R('noun-singular', 'noun-mass', 'character'),
        Rd('CANCER', 0.1, 'noun-singular', 'noun-mass'),
        R('noun-plural', 'noun-mass', 'character')),
    Meme('{} KILLS {}',
        R('noun-singular', 'character'),
        R('noun-plural', 'character')),
    Meme('{0} YOUR CH*RPING {1}. {0} THEM.',
        R('verb-tr-inf'),
        R('noun-plural'))
]

def print_help(f=sys.stdout):
    print("""Usage: automome.py [option] [num]
[option] may be one of the following:
    -h, --help              show this message and exit
    -V, --version           show version and exit
    -u, --unmolpish         include unmolpish memes

[num] is the number of memes to generate; if it is missing or invalid, it is
assumed to be 1.""", file=f)

def print_version():
    print("""automome version 0.1, the Ottish AUTOMEME clone
Copyright (C) 2014 Penguin Development
Inspired by AUTOMEME (http://automeme.net/)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.""")

if __name__ == '__main__':
    n = 1
    unmolpish = False
    opts, args = getopt.gnu_getopt(sys.argv[1:], 'hVu', ('help', 'version',
        'unmolpish'))

    for o, a in opts:
        if o in ('-h', '--help'):
            print_help()
            sys.exit(0)
        elif o in ('-V', '--version'):
            print_version()
            sys.exit(0)
        elif o in ('-u', '--unmolpish'):
            unmolpish = True

    if len(args) > 0:
        n = int(args[0])

    if n < 1:
        n = 1

    if unmolpish:
        memes += memes_unmolpish

    for i in range(n):
        print(str(random.choice(memes)))
